import dataiku
from dataiku.llm.agent_tools import BaseAgentTool
import requests
from dataiku.sql import SelectQuery, toSQL
import re
from typing import Any, Dict, List, Union, Tuple
import logging
from dataiku import SQLExecutor2
import sql_query_utils, sql_tool_prompts

class SQLQueryTool(BaseAgentTool):
    def set_config(self, config, plugin_config):
        self.config = config
        self.api_client = dataiku.api_client()

        datasets_fragment = ""

        for dataset in self.config["datasets"]:

            datasets_fragment += "* Dataset: %s\n\n" % dataset

            dataset = self.api_client.get_default_project().get_dataset(dataset)
            settings = dataset.get_settings()

            # OpenAI tool desc length max is small ...
            for col in settings.schema_columns:
                datasets_fragment += "  * %s" % (col["name"]) # (type: %s)" % (col["name"], col["type"])
                #if "comment" in col:
                #    datasets_fragment += "  - description: %s" % col["comment"]
                datasets_fragment += "\n"

            datasets_fragment += "\n\n"

        self.datasets_fragment = datasets_fragment

    def get_descriptor(self, tool):

        # OpenAI tool desc length max is 1024 :'(
        datasets_fragment = self.datasets_fragment[0:550]

        return {
            "description": """Given a text query explaining the question to respond, this tool will:

1. Determine whether the datasets are likely to be able to respond the question
2. Generate a SQL query 
3. Execute the SQL query
4. Return the results

The question must be respondable by a single SQL query.

If the tool determines that the question cannot be responded, it will return {"status": "cant_respond"}.

Here are the datasets and columns that the tool has access to:

%s""" % datasets_fragment,
            "inputSchema" : {
                "$id": "https://dataiku.com/agents/tools/scrape/input",
                "title": "Input for the SQL querying tool",
                "type": "object",
                "properties" : {
                    "question" : {
                        "type": "string",
                        "description": "The question to respond"
                    }
                },
                "required": ["question"]            
            }
        }

    def invoke(self, input, trace):
        args = input["input"]
        question = args["question"]

        source = {
            "toolCallDescription": "Tried to answer question: %s using SQL datasets" % question,
            "items" : []
        }

        llm = self.api_client.get_default_project().get_llm(self.config["llmId"])

        #########################################################
        # Step 1: Decide whether to use the tool or to bail out
        #########################################################
        with trace.subspan("Deciding whether SQL tool can answer the question") as subspan:
            decision_completion = llm.new_completion()
            decision_completion.with_message(sql_tool_prompts.DECISION_TO_USE_SYSTEM_PROMPT, "system")
            decision_completion.with_message(sql_tool_prompts.DECISION_TO_USE_USER_PROMPT_FMT.format(
                                                    datasets_descriptions=self.datasets_fragment,
                                                    user_question=question), "user")

            decision_response = decision_completion.execute()
            subspan.append_trace(decision_response.trace)

            logging.info("Decision result: %s" % decision_response.text)
            decision_result = decision_response.json
        
            if not "tables_and_columns" in decision_result:
                subspan.outputs["decision"] = "no"

                source["items"].append({
                        "type": "INFO",
                        "textSnippet": "Declined to answer"
                })
                return {
                    "output": { "status": "cant_respond" },
                    "sources":  [source]
                }

            subspan.outputs["decision"] = "yes"

            source["items"].append({
                "type": "INFO",
                "textSnippet": "Decided to use %s" % decision_result["tables_and_columns"]
            });


        #########################################################
        # Step 2: Ask LLM to generate the SQL query as an AST
        #########################################################
        with trace.subspan("Generating SQL query ") as subspan:
            query_building_completion = llm.new_completion()
            query_building_completion.with_message(sql_tool_prompts.QUERY_BUILDING_SYSTEM_PROMPT, "system")
            if "additionalInformation" in self.config:
                query_building_completion.with_message(self.config["additionalInformation"], "system")
            query_building_completion.with_message(sql_tool_prompts.QUERY_BUILDING_USER_PROMPT_FMT.format(
                                                        chosen_tables_and_columns=decision_result["tables_and_columns"],
                                                        decision_to_use_justification=decision_result["justification"],
                                                        user_question=question), "user")

            query_building_response = query_building_completion.execute()
            subspan.append_trace(query_building_response.trace)

            logging.info("Query building result: %s" % query_building_response.text)


        #########################################################
        # Step 3: Expand AST (replace dataset names by table names), transform to SQL query and execute
        #########################################################
        with trace.subspan("Executing SQL query") as subspan:
            sql_query, records, source_items = self.run_sql_query(query_building_response.json)
            subspan.attributes["sqlQuery"] = sql_query

            source["items"].extend(source_items)

        #########################################################
        # Step 4: Ask LLM to synthesize the final answer
        #########################################################
        with trace.subspan("Synthesizing final answer") as subspan:
            final_completion = llm.new_completion()
            final_completion.with_message(sql_tool_prompts.FINAL_ANSWER_SYSTEM_PROMPT, "system")
            if "additionalInformation" in self.config:
                final_completion.with_message(self.config["additionalInformation"], "system")

            final_completion.with_message(sql_tool_prompts.FINAL_ANSWER_USER_PROMPT_FMT.format(
                                                user_question=question,
                                                chosen_tables_and_columns=decision_result["tables_and_columns"],
                                                sql_query=sql_query,
                                                records=records), "user")


            final_response = final_completion.execute()
            subspan.append_trace(final_response.trace)

            logging.info("Final response: %s" % final_response.text)
        
        return { "output" : final_response.text, "sources": [source] }


    def run_sql_query(self, db_query):
        # TODO: Auto retry with correction by the LLM
        
        connection_name = self.config["connection"]
        if connection_name is None:
            raise Exception("A SQL connection selection is required to run a query.")
        
        hard_sql_limit = 200 # TODO

        conn = dataiku.get_connection(connection_name)
        dialect = conn.conn_details["type"]
        executor = SQLExecutor2(connection=connection_name)
        
        select_query = SelectQuery()
        select_list = db_query["selectList"]

        if isinstance(select_list, List):
            select_query._query["selectList"] = select_list
        else:
            select_query._query["selectList"] = [{"expr": {"type": "COLUMN", "name": "*"}}]

        from_ = db_query["from"]
        if isinstance(from_, dict):
            select_query._query["from"] = from_
        else:
            raise Exception("A query must have 'from' statement")

        for key in ["with", "join", "where", "groupBy", "having", "orderBy"]:
            val = db_query[key]
            if isinstance(val, List):
                select_query._query[key] = val

        records = []
        limit = db_query["limit"]
        if isinstance(limit, int):
            if limit < hard_sql_limit:
                select_query._query["limit"] = limit
            else:
                records.append(
                    [
                        {
                            "WARNING!": f"The query exceeded the query limit of {hard_sql_limit}. Some information may be missing. Warn the user!"
                        }
                    ]
                )
                select_query._query["limit"] = hard_sql_limit
        else:
            select_query._query["limit"] = hard_sql_limit

        logging.debug(f"Initial AST: {select_query}")

        sql_query_utils.replace_tables_in_ast(select_query, self.config["datasets"])
    
        logging.debug(f"AST with tables replaced: {select_query}")

        sql_query = toSQL(select_query, dialect=dialect)
        logging.debug(f"SQL query: {sql_query}")


        query_source_item = {
            "type": "GENERATED_SQL_QUERY",
            "performedQuery" : sql_query
        }

        df = executor.query_to_df(query=sql_query)
        df.fillna("", inplace=True)
        records = df.to_dict("records")

        records_source_item = {
            "type": "RECORDS",
            "records": {
                "columns": [str(x) for x in df.columns],
                "data" : [[cell for cell in line] for line in df.values[:10]]
            }
        }

        return sql_query, records[:10], [query_source_item, records_source_item]

        # completion.with_message(sql_tool_prompts.DECISION_TO_USE_SYSTEM_PROMPT, "system")
        # completion.with_message(sql_tool_prompts.DECISION_TO_USE_USER_PROMPT_FMT.format(datasets_descriptions=self.datasets_fragment, user_question=question), "user")

        # self._user_error_prompt = (
        #     """\
        # Correct the following JSON search query response.
        # You must follow the JSON Query Format and use the same keys:
        # - Generate a search query only if there is a need for more information and there were no previous errors.
        # - If no query is needed, all other keys should be null and 'justification' should explain why no query is necessary.
        # - Provide a JSON object only. Do not add any other explanations, decorations, or additional information beyond the a JSON object.
        # - JSON object should have double quotes for keys and values. It is important to follow the JSON format strictly.


        # """
        #     + """
        # # CHAT HISTORY:
        # {chat_history}
        # # --- END OF CHAT HISTORY ---
        # your JSON object with double quotes for keys and values:
        # """
        # )    